import pyarrow as pa
from stonewave.sql.udtfs.protocol.arrow import rename_duplicated_field
from stonewave.sql.udtfs.logger import logger

# left table and right table must have the exact same number of rows
def _lateral_join(left_table, right_table, column_name_qualifier):
    if right_table and right_table.num_columns > 0:
        right_table_schema = right_table.schema
        result_schema = left_table.schema
        arrays = [array for array in left_table]
        for i in range(0, right_table.num_columns):
            field = right_table_schema.field(i)
            field_name = column_name_qualifier + "." + field.name if column_name_qualifier else field.name
            field = pa.field(field_name, field.type)
            field = rename_duplicated_field(result_schema, field)
            result_schema = result_schema.append(field)
            arrays.append(right_table.column(i))
        result_table = pa.RecordBatch.from_arrays(arrays, result_schema.names)
        logger.debug(
            "lateral join left and right tables",
            num_rows=result_table.num_rows,
            num_columns=result_table.num_columns,
        )
        return result_table
    else:
        logger.debug(
            "lateral join with empty right table, return left table as it is",
            left_table=left_table.num_rows,
        )
        return left_table


def _get_cell_value_by_param(batch, row_id, param):
    param_type = param["type"]
    param_value = param["value"]
    if param_type == "literal":
        return param_value
    elif param_type == "column":
        column_name = param_value
        column_index = batch.schema.get_field_index(column_name)
        if column_index == -1:
            # column does not exist, return None
            return None
        column = batch[column_index]
        return column[row_id].as_py()
    return None


def _get_row(batch, row_id, params):
    row = [_get_cell_value_by_param(batch, row_id, param) for param in params]
    return row



import pyarrow as pa
from stonewave.sql.udtfs.logger import logger
from stonewave.sql.udtfs.union_type_resolver import primitive_union_data_type, build_primitive_union_column


class RecordBatchBuilder(object):
    def __init__(self):
        self.columns = {}
        self.schema = {}
        self._row_count = 0

    def column_names(self):
        return self.schema.keys()

    def add_column(self, column_name, data_type=primitive_union_data_type):
        """
        Append a column to the builder. If the column with the same name exists in the builder, it will NOT be added again.
        :param column_name: the name of the column, default value is predefined union data type
                            given a specific data type is recommended so that the function would be more effective.
                            in case the data type is not determined or the column is mixed type,
                            leave the parameter None to make it load union data type.
        :param data_type: pyarrow's field data type, for example, pyarrow::utf8()
        :return: None
        """
        if column_name not in self.columns:
            self.columns[column_name] = []
            self.schema[column_name] = data_type

    def append(self, column, value):
        """
        Append a new value for a given column.
        After appending values for all columns in a row,
        caller needs to call `increase_row_count` to ensure the builder's row_count is increased correctly.
        :param column: the name of the column
        :param value: the new value to be append to the column
        :return:
        """
        self._fill_null_for_all_builders(self._row_count)
        return self.columns.get(column).append(value)

    def extend(self, column, values):
        """
        Append multiple new values for a new given column.
        After appending values for all columns in a row,
        caller needs to call `increase_row_count` to ensure the builder's row_count is increased correctly.
        :param column: the name of the column
        :param values: a list of new values to be appended to the columns
        :return:
        """
        # caller needs to call `increase_row_count` to ensure the builder's row_count is increased correctly
        self._fill_null_for_all_builders(self._row_count)
        return self.columns.get(column).extend(values)

    def num_fields(self):
        """
        :return: the number of fields (columns) currently in this builder
        """
        return len(self.schema)

    def flush(self):
        """
        Flush the contents in the record build to build a record batch.
        return the record batch built from the contents in the builder,
        and return None if the built record batch is empty.
        """
        if self.columns:
            logger.debug(
                "flushing contents from record batch builder",
                columns=len(self.columns),
                row_count=self._row_count,
            )
            self._fill_null_for_all_builders(self._row_count)

            arrow_arrays = [
                build_primitive_union_column(array)
                if self.schema[col].equals(primitive_union_data_type)
                else pa.array(array, self.schema[col])
                for col, array in self.columns.items()
            ]

            batch = pa.RecordBatch.from_arrays(arrow_arrays, list(self.columns.keys()))
            # clear all arrays
            for col in self.columns.keys():
                self.columns[col] = []
            self._row_count = 0
            return batch
        # having null rows only, use empty batch to indicates this
        elif self._row_count > 0:
            self._row_count = 0
            batch = pa.RecordBatch.from_arrays([])
            return batch
        else:
            return None

    def row_count(self):
        """
        Return the number of rows in the builder
        :return:
        """
        return self._row_count

    def increase_row_count(self, increment=1):
        """
        Increase the number of rows in the builder. This is necessary after caller append values for individual column
        by calling `append` or `extend` API in this class.
        :param increment: the number of rows to be increased, by default, this method increases by 1
        :return:
        """
        self._row_count += increment

    def append_null_row(self):
        """
        append a null row to the builder.
        The row count will be automatically maintained.
        :return:
        """
        for array in self.columns.values():
            array.append(None)
        self.increase_row_count()

    def append_row(self, kv_pairs):
        """
        Append a row with as key value pairs, represented by a Python dictionary.
        The key represents the column name and the value represents the column value, which is pyarrow.utf8() type
        :param kv_pairs: the row's field names and values
        :return:
        """
        self._fill_null_for_all_builders(self._row_count)
        for key, value in kv_pairs.items():
            # TODO: we don't support creating multi value field yet in python
            self.add_column(key, pa.utf8())
            array = self._get_column_array(key)
            array.append(value)
        self.increase_row_count()

    def _fill_null_for_all_builders(self, new_row_count):
        for array in self.columns.values():
            null_count = new_row_count - len(array)
            for _ in range(0, null_count):
                array.append(None)

    def _get_column_array(self, column):
        return self.columns.get(column)



from stonewave.sql.udtfs.base_function import BaseFunction
import pyarrow as pa


class SummarizeFunction(BaseFunction):
    def __init__(self):
        self.tables = []

    def get_name(self):
        return "summarize"

    def process(self, row_writer, row_idx, args):
        assert len(args) > 0
        batch = args[0]
        if batch is not None:
            table = pa.Table.from_batches([batch])
            # FIXME: the current implementation caches all batches, which is not necessary
            self.tables.append(table)
        else:
            self.summarize(row_writer)

    def summarize(self, row_writer):
        if self.tables:
            table = pa.concat_tables(self.tables, promote=True)
            df = table.to_pandas()
            desc_df = df.describe(datetime_is_numeric=True)
            # the df is transposed because it contains mixed type column, which is not allowed in arrow
            desc_df = desc_df.transpose()
            desc_df.insert(0, "fields", desc_df.index)
            desc_table = pa.Table.from_pandas(desc_df, preserve_index=False)
            batches = desc_table.to_batches()
            row_writer.batch_iterator = iter(batches)
        else:
            return



from stonewave.sql.udtfs.record_batch_builder import RecordBatchBuilder


class TableRowWriter(object):
    """
    Use the row writer to write results produced from a table function. There are two major ways to write results:
    1) get `record_batch_builder` and write the produced rows row by row. See `RecordBatchBuilder`'s API for details.
    2) set `batch_iterator` that will produce a list of record batches.
    The `batch_iterator` MUST be a generator that calling its next method will return a pyarrow's RecordBatch
    """

    def __init__(self, flushing_batch_size=8 * 1024):
        self._flushing_batch_size = flushing_batch_size
        self._record_batch_builder = RecordBatchBuilder()
        self._batch_iterator = None
        self._current_builder_row_count = 0
        self._last_flushed_batch_size = 0
        self._new_rows_count = 0

    @property
    def record_batch_builder(self):
        return self._record_batch_builder

    @record_batch_builder.setter
    def record_batch_builder(self, value):
        self._record_batch_builder = value

    @property
    def batch_iterator(self):
        return self._batch_iterator

    @batch_iterator.setter
    def batch_iterator(self, value):
        self._batch_iterator = value

    def builder_row_count(self):
        return self.record_batch_builder.row_count()

    def is_not_empty(self):
        return self.batch_iterator or self.record_batch_builder.row_count() > 0

    def append_row(self, row):
        self._record_batch_builder.append_row(row)

    def append_rows(self, rows):
        for row in rows:
            self._record_batch_builder.append_row(row)

    def flush(self, forced=False):
        self._before_flush()
        flushed_batch = None
        if self.batch_iterator:
            batch = self.record_batch_builder.flush()
            if batch is not None:
                flushed_batch = batch
            else:
                try:
                    flushed_batch = next(self.batch_iterator)
                except StopIteration:
                    self.batch_iterator = None
                    return None
        elif self._is_over_flushing_size() or forced and self.record_batch_builder.row_count() > 0:
            flushed_batch = self.record_batch_builder.flush()

        self._last_flushed_batch_size = flushed_batch.num_rows if flushed_batch else 0
        self._after_flush()
        return flushed_batch

    ##################################################
    # APIs below are internal and is subject to change
    ##################################################
    @property
    def new_rows_count(self):
        """
        :return: the number of new rows after processing a new row or flushing results
        """
        return self._new_rows_count

    def _before_process(self):
        self._current_builder_row_count = self.record_batch_builder.row_count()

    def _after_process(self):
        self._new_rows_count = self.record_batch_builder.row_count() - self._current_builder_row_count

    def _before_flush(self):
        self._current_builder_row_count = self.record_batch_builder.row_count()

    def _after_flush(self):
        self._new_rows_count = (
            self.record_batch_builder.row_count() + self._last_flushed_batch_size - self._current_builder_row_count
        )

    def _is_over_flushing_size(self):
        return self.record_batch_builder.row_count() >= self._flushing_batch_size



import os
import pathlib
from stonewave.sql.udtfs.logger import logger
from stonewave.sql.udtfs.constants import STONEWAVE_HOME


class BaseDataloader(object):
    def __init__(self):

        self.external_data_dir = os.path.join(STONEWAVE_HOME, "var", "external_data")

    def set_path(self, path):
        logger.debug(
            "executing set_path in base dataloader",
            path=path,
            external_data_dir=self.external_data_dir,
        )
        if os.path.isabs(path):
            abs_path = path
        else:
            abs_path = os.path.join(self.external_data_dir, path)

        data_path = os.path.realpath(abs_path)
        external_data_length = len(self.external_data_dir)

        if len(data_path) < external_data_length or data_path[:external_data_length] != self.external_data_dir:
            raise Exception("failed to load path {}: no data or directory exists in external data path.".format(path))

        data_path = pathlib.Path(data_path)

        if not data_path.exists():
            raise Exception("failed to load path {}: no such data or directory.".format(data_path))

        self.path = data_path

    def get_path(self):
        return self.path



import pyarrow as pa
import json
from sql_metadata import Parser
from stonewave.sql.udtfs.base_function import BaseFunction, udtf
import re

STRING_LIST_TYPE = pa.list_(pa.utf8())


def _prefix_table_name(table_name, column_name):
    return "{}.{}".format(table_name, column_name)


def _try_with_default_value(obj, property, default_value):
    try:
        return getattr(obj, property)
    except Exception as e:
        return default_value


def _is_single_table_without_cte(tables, with_queries):
    return len(tables) == 1 and not with_queries


def _pre_process_sql(sql):
    return sql.replace("?", "1")


def _dedup_and_fix_tables(tables):
    tables = sorted(set(tables))
    # sql-metadata doesn't parse `FROM (WITH cte (...))` correctly
    for t in tables:
        if "WITH" == t.upper():
            tables.remove(t)
    return tables


def _parse_query(sql):
    sql = _pre_process_sql(sql)
    parser = Parser(sql)
    query_type = _try_with_default_value(parser, "query_type", "unknown")
    tables = _dedup_and_fix_tables(_try_with_default_value(parser, "tables", []))
    with_queries = _try_with_default_value(parser, "with_queries", {})
    sub_queries = _try_with_default_value(parser, "subqueries", {})
    columns = _try_with_default_value(parser, "columns", [])
    table_alias = _try_with_default_value(parser, "tables_aliases", {})

    if _is_single_table_without_cte(tables, with_queries):
        columns = [_prefix_table_name(tables[0], c) if "." not in c else c for c in columns]

    for _, query in with_queries.items():
        columns_in_cte, _, _ = _parse_query(query)
        columns += columns_in_cte

    for _, query in sub_queries.items():
        columns_in_sub_query, _, _ = _parse_query(query)
        columns += columns_in_sub_query

    return (columns, tables, query_type)


def _is_table_function(sql, table):
    return f"{table}(" in sql


# TODO: sql-metadata doesn't support table function yet,
# https://github.com/macbre/sql-metadata/issues/276
# here is a hacky approach by replacing `my_func()` with `my_func`
def _pre_process_table_function(sql):
    parser = Parser(sql)
    tables = _dedup_and_fix_tables(_try_with_default_value(parser, "tables", []))
    for table in tables:
        if _is_table_function(sql, table):
            pattern = re.compile(f"{table}\\((.*?)\\)")
            sql = re.sub(pattern, table, sql)

    return sql


@udtf(is_parser=True)
class ParseSqlFunction(BaseFunction):
    def __init__(self):
        pass

    def get_name(self):
        return "parse_sql"

    def process(self, row_writer, row_idx, args):
        sql = args[0]
        sql = _pre_process_table_function(sql)

        record_batch_builder = row_writer.record_batch_builder
        record_batch_builder.add_column("columns", STRING_LIST_TYPE)
        record_batch_builder.add_column("tables", STRING_LIST_TYPE)
        record_batch_builder.add_column("query_type", pa.utf8())

        columns, tables, query_type = _parse_query(sql)

        columns = sorted(set(columns))
        record_batch_builder.append("columns", columns)
        record_batch_builder.append("tables", tables)

        record_batch_builder.append("query_type", query_type)
        record_batch_builder.increase_row_count()



from transitions import Machine
from stonewave.sql.udtfs.protocol import ipc
from stonewave.sql.udtfs.constants import PROCESS_ERROR_MSG_TEMPLATE
from stonewave.sql.udtfs.table_row_writer import TableRowWriter
from stonewave.sql.udtfs.logger import logger


class EvalFunctionFsm(object):
    states = ["start", "wait_for_next", "end"]

    def __init__(self, func, batch_sender):
        self.func = func
        self._batch_sender = batch_sender
        self._row_writer = TableRowWriter()
        self._is_evaluated = False

        self.machine = Machine(model=self, states=EvalFunctionFsm.states, initial="start")

        self.machine.add_transition(
            trigger="eval",
            source="start",
            dest="wait_for_next",
            after="eval_params",
        )

        self.machine.add_transition(
            trigger="next",
            source="wait_for_next",
            dest="wait_for_next",
            after="send_next_batch",
        )

        self.machine.add_transition(trigger="end", source="*", dest="end", before="end_evaluation")

    def eval_params(self, params, respond):
        self.send_next_batch(params, respond)

    def send_next_batch(self, params, respond):
        if not self._is_evaluated:
            self.func.initialize(self._row_writer)
            func_name = self.func.get_name()
            try:
                self.func.process(self._row_writer, 0, params)
            except Exception as e:
                logger.error(
                    "failed to evaluate table function",
                    function=func_name,
                    params=str(params),
                    error=PROCESS_ERROR_MSG_TEMPLATE.format(func_name, str(e)),
                )
                raise Exception(PROCESS_ERROR_MSG_TEMPLATE.format(func_name, str(e)))
            finally:
                self._is_evaluated = True

        batch = self._row_writer.flush(True)
        if batch is not None:
            self._batch_sender.send(batch, respond)
            return
        else:
            self.end(params, respond)

    def end_evaluation(self, params, respond):
        respond("finish")



import multiprocessing as mp
from threading import Semaphore, Lock
from stonewave.sql.udtfs.logger import logger


def singleton(cls, *args, **kw):
    instances = {}

    def _singleton(*args, **kw):
        if cls not in instances:
            instances[cls] = cls(*args, **kw)
        return instances[cls]

    return _singleton


@singleton
class TaskManager(object):
    # ideally the pool_size can be controlled by a config variable
    def __init__(self, pool_size=mp.cpu_count()):
        self._reset(pool_size=pool_size)

    ## add a reset function used in unit test for this singleton object, mock reinitilization of TaskManager
    def _reset(self, pool_size=mp.cpu_count()):
        if hasattr(self, "_pool"):
            self._pool.close()
            self._pool.terminate()
        logger.info("init worker pool", pool_size=pool_size)
        # TODO change max_limit to a configuarable constant
        self.max_limit = 32
        self.max_pool_size = pool_size
        self._pool = mp.Pool(processes=pool_size, maxtasksperchild=1)
        self.workers = Semaphore(pool_size)
        # ensures the semaphore is not replaced while used
        self.workers_mutex = Lock()
        self._pipe_map = {}

    def get_pool(self):
        return self._pool

    def get_map(self):
        return self._pipe_map

    def change_pool_size(self, new_size):
        """Set the Pool to a new size."""
        logger.info("change task manager pool size")
        with self.workers_mutex:
            if new_size > self.max_limit or new_size <= 0:
                logger.error("new pool size not in the range of 0 and cpu count size")
                return False
            current_value = self.workers._value
            diff_value = new_size - self.max_pool_size
            value = current_value + diff_value
            if value < 0:
                logger.error(
                    "pool workers still running, semaphore value can not be negative", changed_semaphore_value=value
                )
                return False
            self.workers = Semaphore(value)
            self.max_pool_size = new_size
            return True
        return False

    def new_task(self, task, args):
        """Start a new task, blocks if queue is full."""
        with self.workers_mutex:
            if not self.workers.acquire(blocking=False):
                return False

        result = self._pool.apply_async(task, args=args, callback=self.task_done)
        return result

    def task_done(self, *args, **kwargs):
        """Called once task is done, releases the queue is blocked."""
        with self.workers_mutex:
            logger.debug("function execution task done")
            self.workers.release()



from stonewave.sql.udtfs.base_function import BaseFunction
import pyarrow as pa


class UnpivotTableFunction(BaseFunction):
    def __init__(self):
        self.tables = []
        self.x_field = None
        self.y_name_field = None
        self.y_data_field = None

    def get_name(self):
        return "unpivot_table"

    def process(self, row_writer, row_idx, args):
        if self.x_field == None and len(args) < 4:
            raise Exception(
                "Table function 'unpivot_table' parameter invalid: "
                "parameter should be (data_set_name, index_field, "
                "column_name, column_value_name)"
            )
        batch = args[0]
        if batch is not None:
            self.x_field = args[1]
            self.y_name_field = args[2]
            self.y_data_field = args[3]
            table = pa.Table.from_batches([batch])
            # FIXME: the current implementation caches all batches, which is not necessary
            self.tables.append(table)
        else:
            self.unpivot(row_writer)

    def unpivot(self, row_writer):
        if self.tables:
            table = pa.concat_tables(self.tables, promote=True)
            df = table.to_pandas()
            unpvt = df.melt(
                id_vars=self.x_field,
                var_name=self.y_name_field,
                value_name=self.y_data_field,
            )
            unpvt_table = pa.Table.from_pandas(unpvt, preserve_index=False)
            batches = unpvt_table.to_batches()
            row_writer.batch_iterator = iter(batches)
        else:
            return



from stonewave.sql.udtfs.protocol import ipc
from stonewave.sql.udtfs.logger import logger


class SharedMemoryRecordBatchSender(object):
    def send(self, batch, respond):
        logger.debug(
            "send record batch via shared memory",
            num_rows=batch.num_rows,
            num_columns=batch.num_columns,
        )
        shm_info = ipc.share_record_batch(batch)
        respond({"shm_name": shm_info[0], "shm_size": shm_info[1]})



def udtf(is_lookup=False, is_parser=False, not_producing_multi_rows=False):
    """
    @param is_parser: `is_parser=True` indicates this table function parses its argument and produces key value pairs as a new row. It should NOT write more than 1 row for each input if specified. Specifying it to be True helps performance.
    @param is_lookup: `is_lookup=True` indicates this table function uses its argument as lookup key and produces key value pairs as a new row. It should NOT write more than 1 row for each input if specified. Specifying it to be True helps performance.
    @param not_producing_multi_rows: `not_producing_multi_rows=True` indicates this table function produces no more than one row. It should NOT write more than 1 row for each input if specified.
    """

    def decorate_with_udtf(func_class):
        func_class.is_parser = is_parser
        func_class.is_lookup = is_lookup
        func_class.not_producing_multi_rows = is_lookup or is_parser or not_producing_multi_rows
        return func_class

    return decorate_with_udtf


class BaseFunction(object):
    def get_name(self):
        """
        :return: the name of the table function
        """
        return "base"

    def initialize(self, table_row_writer):
        """
        This method will be called once for every batch in the input table with function applied
        :param table_row_writer: the row writer for writing produced results
        :return: None
        """
        pass

    def process(self, table_row_writer, row_idx, row):
        """
        This method is used to process the current row of input, and it will be called once for every row in an input table.
        :param table_row_writer: Use the `table_row_writer` to write the produced rows into the result table.
        See the table_row_writer API for details.
        :param row_idx: the row index for the current row and current batch
        :param row: a list containing all of the parameters from the current row
        :return: None
        """
        pass


BaseFunction.is_parser = False
BaseFunction.is_lookup = False
BaseFunction.not_producing_multi_rows = False



import pyarrow as pa
from datetime import datetime

_primitive_data_types = [
    (type(None), pa.null()),
    (int, pa.int64()),
    (float, pa.float64()),
    (bool, pa.bool_()),
    (str, pa.utf8()),
    (datetime, pa.timestamp("us")),
    (list, pa.list_(pa.utf8())),
]
primitive_union_data_type = pa.union(
    [pa.field(str(idx), dt[1]) for idx, dt in enumerate(_primitive_data_types)], "dense"
)


def build_primitive_union_column(column_values):
    type_ids = []
    offsets = []
    child_arrays = [] * 7
    for idx, type_tuple in enumerate(_primitive_data_types):
        child_array = []
        offset_starts = 0
        parsed_indices = []
        for val_idx, value in enumerate(column_values):
            if type(value) == type_tuple[0]:
                type_ids.append(idx)
                offsets.append(offset_starts)
                child_array.append(value)
                offset_starts += 1
                parsed_indices.append(val_idx)
        for i in reversed(parsed_indices):
            column_values.pop(i)
        child_arrays.append(pa.array(child_array, type=type_tuple[1]))
    return pa.UnionArray.from_dense(
        pa.array(type_ids, type=pa.int8()), pa.array(offsets, type=pa.int32()), child_arrays
    )



import json
import os
import argparse
import sys
import pkgutil
import toml
import shutil
import time
import subprocess
from stonewave.sql.udtfs.function_executor import execute
from stonewave.sql.udtfs.constants import (
    USER_DEFINED_TABLE_FUNCTIONS_PATH,
    USER_DEFINED_TABLE_FUNCTION_INFO_FILE,
    SIGNATURE_LIST,
)
from stonewave.sql.udtfs.logger import logger
import stonewave.sql.udtfs.functions as built_in_funcs
from wheel_filename import parse_wheel_filename, InvalidFilenameError
from stonewave.sql.udtfs.test_utility import check_expected_parameters_list
from stonewave.sql.udtfs.load_function import load_function_by_name
from stonewave.sql.udtfs.version import sql_udtfs_version


def _list_udtfs_from_path(path, udtfs_list):
    try:
        for importer, modname, _ in pkgutil.iter_modules(path):
            info_path = os.path.join(importer.path, modname, USER_DEFINED_TABLE_FUNCTION_INFO_FILE)
            if not os.path.exists(info_path):
                logger.error("can not load function information", function_name=modname, info_path=info_path)
                continue
            info_toml = toml.load(info_path)
            udtfs_list[modname] = info_toml[SIGNATURE_LIST]
    except Exception as e:
        logger.error("load udtfs error", path=path, error=str(e))


def list_udtfs():
    if not os.path.exists(USER_DEFINED_TABLE_FUNCTIONS_PATH):
        os.mkdir(USER_DEFINED_TABLE_FUNCTIONS_PATH)
    udtfs_list = {}
    _list_udtfs_from_path(built_in_funcs.__path__, udtfs_list)
    _list_udtfs_from_path([USER_DEFINED_TABLE_FUNCTIONS_PATH], udtfs_list)
    return udtfs_list


def list_udtfs_cmd(args):
    udtfs_list = list_udtfs()
    if args.output_path:
        with open(args.output_path, "w") as f:
            f.write(json.dumps(udtfs_list, indent=4))
    else:
        print(json.dumps(udtfs_list, indent=4))


def _validate_function(func_name):
    sys.path.append("/tmp")
    sys.path.append("/tmp/{}".format(func_name))
    func = load_function_by_name(func_name)
    if func is None:
        raise Exception("Can not find method implements from BaseFunction in stonewave.sql.udtfs.base_function")
    func_dir = func().__dir__()
    if "get_name" not in func_dir or "process" not in func_dir:
        raise Exception("Invalid function class, please implement get_name(self) and process(self, )")
    os.system("rm -rf /tmp/{}/**/__pycache__".format(func_name))


def _read_signature_list(path):
    info_toml = toml.load(path)
    return info_toml[SIGNATURE_LIST]


def _raise_register_exception(temporary_working_dir, msg):
    shutil.rmtree(temporary_working_dir)
    raise Exception(msg)


def register_udtf(func_name, func_path, update=False):
    temporary_working_dir = "/tmp/{}_{}".format(func_name, int(round(time.time() * 1000)))
    # install package to temperary directory, to avoid installing an invalid function
    installation = subprocess.run(["pip", "install", func_path, "-t", temporary_working_dir], stderr=subprocess.PIPE)
    # noted that pip always throw some warnings, which will make return code as non zero
    # here should not use installation.return_code to check
    if not os.path.exists(temporary_working_dir):
        raise Exception("Failed to install python package: {}".format(installation.stderr.decode("utf-8")))
    project_name = None
    try:
        # func_path is wheel package file path
        project_name = parse_wheel_filename(os.path.basename(func_path)).project
    except InvalidFilenameError:
        # func_path is pypi package name
        project_name = func_path.replace("-", "_")
    try:
        filelist = os.listdir(os.path.join(temporary_working_dir, project_name))
        for file in filelist:
            src = os.path.join(temporary_working_dir, project_name, file)
            dst = os.path.join(temporary_working_dir, file)
            shutil.move(src, dst)
        shutil.rmtree(os.path.join(temporary_working_dir, project_name))
    except Exception as e:
        logger.error("project path name is not as same as package name", error_message=str(e))
        _raise_register_exception("Please make sure project path name is as same as package name")
    if not update:
        info_path = os.path.join(temporary_working_dir, USER_DEFINED_TABLE_FUNCTION_INFO_FILE)
        if not os.path.exists(info_path):
            logger.error(
                "can not load function information",
                function_name=func_name,
                info_path=USER_DEFINED_TABLE_FUNCTION_INFO_FILE,
            )
            _raise_register_exception(temporary_working_dir, "Please include info.toml for function information")
        try:
            check_expected_parameters_list(_read_signature_list(info_path))
        except Exception as e:
            _raise_register_exception(temporary_working_dir, str(e))
    else:
        existing_info_path = os.path.join(
            USER_DEFINED_TABLE_FUNCTIONS_PATH, func_name, USER_DEFINED_TABLE_FUNCTION_INFO_FILE
        )
        if not os.path.exists(existing_info_path):
            _raise_register_exception(
                temporary_working_dir, "Should not update build-in table function '{}'".format(func_name)
            )
        existing_signature_list = _read_signature_list(existing_info_path)
        new_info_path = os.path.join(temporary_working_dir, USER_DEFINED_TABLE_FUNCTION_INFO_FILE)
        if not os.path.exists(new_info_path):
            logger.error(
                "can not load function information",
                function_name=func_name,
                info_path=USER_DEFINED_TABLE_FUNCTION_INFO_FILE,
            )
            _raise_register_exception(temporary_working_dir, "Please include info.toml for function information")
        new_signature_list = _read_signature_list(new_info_path)
        if existing_signature_list != new_signature_list:
            _raise_register_exception(temporary_working_dir, "Should not update table function signature list")

    _validate_function(os.path.basename(temporary_working_dir))
    shutil.move(temporary_working_dir, os.path.join(USER_DEFINED_TABLE_FUNCTIONS_PATH, func_name))


def register_udtf_cmd(args):
    func_name = args.function_name
    func_path = args.function_path
    update = args.update
    try:
        register_udtf(func_name, func_path, update)
    except Exception as e:
        print(str(e))
        exit(-1)


def exec_udtf_cmd(args):
    execute(args.function_name, input, sys.stdout)


def remove_udtf(function_name):
    os.system("rm -rf {}/{}".format(USER_DEFINED_TABLE_FUNCTIONS_PATH, function_name))


def remove_udtf_cmd(args):
    remove_udtf(args.function_name)


def execute_command():
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description="""
        Command sw_py_udtf
        Usage Examples:

        1) exec
        $ sw_py_udtf exec --function-name <name>
        Execute function, please integrate to stonewave service. Function name is required.
        2) list 
        $ sw_py_udtf list --output-path <path>
        List all accessiable python table functions. Output path is not required. If no
        output path, with list all table functions to stdout.
        3) register
        $ sw_py_udtf register --function-name <name> --function-path <path>
        Function name and function path are required. currently only support file system path
        for user defined table function python wheel package
        4) remove
        $ sw_py_udtf exec --function-name remove
        Remove table function. Function name is required
        """,
    )
    parser.add_argument("-V", "--version", action="version", version=sql_udtfs_version())
    subparsers = parser.add_subparsers()

    exec_subparser = subparsers.add_parser("exec")
    exec_subparser.add_argument(
        "-n",
        "--function-name",
        required=True,
        type=str,
        help="user defined table function name",
    )
    exec_subparser.set_defaults(callback=exec_udtf_cmd)

    list_subparser = subparsers.add_parser("list")
    list_subparser.add_argument(
        "-o",
        "--output-path",
        default="",
        type=str,
        help="output udtfs list to specific path",
    )
    list_subparser.set_defaults(callback=list_udtfs_cmd)

    register_subparser = subparsers.add_parser("register")
    register_subparser.add_argument(
        "-n",
        "--function-name",
        required=True,
        type=str,
        help="user defined table function name",
    )
    register_subparser.add_argument(
        "-p",
        "--function-path",
        required=True,
        type=str,
        help="user defined table function package path",
    )
    register_subparser.add_argument(
        "-u",
        "--update",
        action="store_true",
        help="flag to update existing user defined table function",
    )
    register_subparser.set_defaults(callback=register_udtf_cmd)

    remove_subparser = subparsers.add_parser("remove")
    remove_subparser.add_argument(
        "-n",
        "--function-name",
        required=True,
        type=str,
        help="user defined table function name",
    )
    remove_subparser.set_defaults(callback=remove_udtf_cmd)

    args = parser.parse_args()
    args.callback(args)


if __name__ == "__main__":
    execute_command()



import logging.config
import structlog
import os
import pathlib
from structlog.contextvars import (
    bind_contextvars,
    clear_contextvars,
    merge_contextvars,
    unbind_contextvars,
)


LOG_LEVEL = os.getenv("SW_LOG_LEVEL") or "INFO"
LOG_LEVEL = str(LOG_LEVEL).upper()
timestamper = structlog.processors.TimeStamper(fmt="%Y-%m-%d %H:%M:%S")
log_file_dir = os.path.join(os.getenv("STONEWAVE_HOME", "/tmp"), "logs", "py_table_funcs")
pathlib.Path(log_file_dir).mkdir(parents=True, exist_ok=True)


def get_logger(logger_id=None):
    if not os.path.exists(log_file_dir):
        os.mkdir(log_file_dir)
    if logger_id is None:
        logger_config = get_logger_config()
    else:
        log_file = os.path.join(log_file_dir, "py_table_funcs.{}.log".format(logger_id))
        logger_config = get_logger_config(log_file)
    logging.config.dictConfig(logger_config)
    structlog.configure(
        processors=[
            merge_contextvars,
            structlog.stdlib.add_log_level,
            structlog.stdlib.PositionalArgumentsFormatter(),
            timestamper,
            structlog.processors.StackInfoRenderer(),
            structlog.processors.format_exc_info,
            structlog.stdlib.ProcessorFormatter.wrap_for_formatter,
        ],
        context_class=dict,
        logger_factory=structlog.stdlib.LoggerFactory(),
        wrapper_class=structlog.stdlib.BoundLogger,
        cache_logger_on_first_use=True,
    )

    logger = structlog.get_logger()
    bind_contextvars(pid=os.getpid())
    return logger


def get_logger_config(log_file=os.path.join(log_file_dir, "py_table_funcs.{}.log".format(os.getpid() % 1000))):
    pre_chain = [
        # Add the log level and a timestamp to the event_dict if the log entry
        # is not from structlog.
        structlog.stdlib.add_log_level,
        timestamper,
    ]

    logger_config = {
        "version": 1,
        "disable_existing_loggers": False,
        "formatters": {
            "plain": {
                "()": structlog.stdlib.ProcessorFormatter,
                "processor": structlog.dev.ConsoleRenderer(colors=False),
                "foreign_pre_chain": pre_chain,
            },
        },
        "handlers": {
            "file": {
                "level": LOG_LEVEL,
                "class": "logging.handlers.RotatingFileHandler",
                "filename": log_file,
                "formatter": "plain",
                "maxBytes": 2097152,  # 2MB
                "backupCount": 1,
            },
        },
        "loggers": {
            "": {
                "handlers": ["file"],
                "level": LOG_LEVEL,
                "propagate": True,
            },
        },
    }
    return logger_config


logger = get_logger()



from stonewave.sql.udtfs.protocol.shared_memory_record_batch_reader import (
    SharedMemoryRecordBatchReader,
)


def read_record_batch(shm_info):
    shm_name, shm_size = shm_info["shm_name"], shm_info["shm_size"]
    batch = None
    # an empty shm_name indicates there is no batch passed into the function
    if shm_name:
        reader = SharedMemoryRecordBatchReader(shm_name, shm_size)
        batch = reader.read()
    return batch



import pyarrow as pa
from multiprocessing.resource_tracker import unregister

from stonewave.sql.udtfs.table_row_writer import TableRowWriter
from stonewave.sql.udtfs.logger import logger
from stonewave.sql.udtfs.protocol.fsm.shared_record_batch_reader import read_record_batch
from stonewave.sql.udtfs.protocol.fsm.apply import _lateral_join, _get_row
from stonewave.sql.udtfs.protocol.fsm.apply.equal_length_apply import _equal_length_apply
from stonewave.sql.udtfs.protocol.fsm.apply.apply_left_table_builder import (
    ApplyLeftTableBuilder,
)


class ApplyOperator(object):
    def __init__(self, func, params, flushing_batch_size=8 * 1024):
        self._func = func
        self._params = params
        self._flushing_batch_size = flushing_batch_size

    def execute(self):
        """
        params will contain the info about left table for apply operation, it is a shared memory name and size
        params will contains other parameters when calling the APPLY
        For example, `APPLY parse_regex(_message, regex_pattern)`, the params will be: [{shm_name, shm_size}, is_outer_apply, column_name_qualifier, {expr}, {expr}]
        """
        assert len(self._params) > 1
        shm_info = self._params[0]
        left_table = read_record_batch(shm_info)
        is_outer_apply = self._params[1]
        column_name_qualifier = self._params[2]
        func_params = self._params[3:]

        if is_outer_apply and self._func.__class__.not_producing_multi_rows:
            yield _equal_length_apply(
                self._func,
                left_table,
                column_name_qualifier,
                func_params,
            )
        else:
            yield from _generic_apply_func(
                self._func,
                left_table,
                is_outer_apply,
                column_name_qualifier,
                func_params,
                self._flushing_batch_size,
            )


def _generic_apply_func(
    func,
    left_table,
    is_outer_apply,
    column_name_qualifier,
    func_params,
    flushing_batch_size,
):
    left_table_builder = ApplyLeftTableBuilder(left_table)
    rwriter = TableRowWriter(flushing_batch_size)

    for i in range(0, left_table.num_rows):
        yield from _process_row(
            i,
            func,
            is_outer_apply,
            func_params,
            left_table,
            left_table_builder,
            rwriter,
            column_name_qualifier,
        )

    # if there is remaining non flushed contents in the right table
    if rwriter.is_not_empty():
        logger.debug(
            "force flushing the remaining right table",
            batch_iterator=rwriter.batch_iterator,
            builder_row_count=rwriter.builder_row_count(),
        )
        right_table = rwriter.flush(True)
        if right_table:
            logger.debug("remaining right table flushed", right_table_size=right_table.num_rows)
        yield _lateral_join(
            left_table_builder.flush(),
            right_table,
            column_name_qualifier,
        )


def _process_row(
    i,
    func,
    is_outer_apply,
    func_params,
    left_table,
    lbuilder,
    rwriter,
    column_name_qualifier,
):
    initial_total_repeats = lbuilder.get_total_repeats()

    rwriter._before_process()
    try:
        if i == 0:
            func.initialize(rwriter)
        row = _get_row(left_table, i, func_params)
        func.process(rwriter, i, row)

        rwriter._after_process()
        lbuilder.repeat(i, rwriter.new_rows_count)

        while True:
            right_table = rwriter.flush(forced=False)
            if _has_only_null_rows(right_table):
                yield lbuilder.flush()
                continue
            lbuilder.repeat(i, rwriter.new_rows_count)

            if right_table:
                joined_table = _lateral_join(
                    lbuilder.flush(),
                    right_table,
                    column_name_qualifier,
                )
                yield joined_table
            else:
                break
    except Exception as e:
        logger.error(
            "failed to process row",
            function=func.get_name(),
            args=func_params,
            error=str(e),
        )

    # no right table row is produced after processing this left table row
    if is_outer_apply and lbuilder.get_total_repeats() == initial_total_repeats:
        lbuilder.repeat(i, 1)
        rwriter.record_batch_builder.append_null_row()
    logger.debug("row processed", row_idx=i, left_table_length=lbuilder.get_total_repeats())


# when a number of null rows are appended to the builder, the flushed right table will be an empty batch
# this could happen when apply is outer apply, and the previous row(s) produces empty rows
def _has_only_null_rows(right_table):
    return right_table is not None and right_table.num_rows == 0



import importlib
import inspect
from stonewave.sql.udtfs.base_function import BaseFunction
from stonewave.sql.udtfs.logger import logger


def _find_base_function_from_lib(lib):
    for mod in dir(lib):
        method = eval("lib." + mod)
        if inspect.isclass(method):
            if issubclass(method, BaseFunction):
                if not issubclass(BaseFunction, method):
                    return method
    return None


def load_function_by_name(name):
    try:
        # built-in table functions
        lib = importlib.import_module("stonewave.sql.udtfs.functions.{}".format(name))
        logger.info("load built-in table function", function_name=name)
        return _find_base_function_from_lib(lib)
    except:
        try:
            # user defined table functions
            # functions path has been append to sys.path when starting the executor

            lib = importlib.import_module(name)
            logger.info("load user defined table function", function_name=name)
            return _find_base_function_from_lib(lib)
        except:
            try:
                importlib.invalidate_caches()
                lib = importlib.import_module(name)
                logger.info("clean cache and load user defined table function", function_name=name)
                return _find_base_function_from_lib(lib)
            except Exception as e:
                logger.error("function not found", function_name=name, error=str(e))
                return None



import importlib.metadata

package_metadada = importlib.metadata.metadata("stonewave-sql-udtfs")
# info from pyproject.toml's `version`
VERSION = package_metadada.get("Version")


def sql_udtfs_version():
    return VERSION



from fastapi import FastAPI, HTTPException
import multiprocessing as mp
from pydantic import BaseModel, Json
from typing import Optional
from stonewave.sql.udtfs.task_manager import TaskManager
import uvicorn
import asyncio
import stonewave.sql.udtfs.function_executor as function_executor
from stonewave.sql.udtfs.logger import logger, get_logger_config
import json

app = FastAPI()


class RequestObj(BaseModel):
    func_name: Optional[str] = None
    method: str
    params: Json

    class Config:
        schema_extra = {"example": {"func_name": "faker", "method": "eval", "params": {}}}


@app.post("/table-functions/execs")
async def create(req: RequestObj):
    try:
        func_name = req.func_name
        parent_conn, child_conn = mp.Pipe()
        task_manager = TaskManager()
        worker_value = task_manager.workers._value
        print(worker_value)
        while True:
            if task_manager.new_task(
                function_executor.execute_worker,
                (
                    func_name,
                    child_conn,
                    worker_value,
                ),
            ):
                break
            else:
                print("waiting at worker value {}".format(worker_value))
                await asyncio.sleep(0.2)

        parent_conn.send(req)
        response = parent_conn.recv()
        res_json = json.loads(response)
        execution_id = res_json.get("execution_id")
        if res_json.get("result") == "finish":
            parent_conn.close()
            logger.info("finish execution in create", execution_id=execution_id, req=str(req))
            return res_json
        queue_map = task_manager.get_map()
        if execution_id is None or queue_map.get(execution_id) is not None:
            raise HTTPException(status_code=500, detail="queue map execution_id not valid")
        logger.info("start execution", execution_id=execution_id, req=str(req), worker=worker_value)
        queue_map[execution_id] = parent_conn
        return res_json
    except HTTPException as e:
        print("error occured {}".format(str(e)))
        raise e
    except Exception as e:
        print("error occured {}".format(str(e)))
        raise HTTPException(status_code=500, detail=str(e))


@app.post("/table-functions/execs/{execution_id}")
async def receive_req(req: RequestObj, execution_id: str):
    try:
        execution_id = execution_id
        task_manager = TaskManager()
        queue_map = task_manager.get_map()
        if execution_id is not None:
            parent_conn = queue_map.get(execution_id)
            if parent_conn is None:
                raise HTTPException(status_code=500, detail="queue map execution_id not valid")
            parent_conn.send(req)
            response = parent_conn.recv()
            res_json = json.loads(response)
            if res_json["result"] == "finish":
                pool_instance = TaskManager()
                queue_map.pop(execution_id)
                parent_conn.close()
                logger.info(
                    "finish execution, remaining number of workers: {}".format(pool_instance.workers._value),
                    execution_id=execution_id,
                    req=str(req),
                )
            return res_json
        else:
            raise HTTPException(status_code=400, detail="no execution_id provided")
    except HTTPException as e:
        print("error occured {}".format(str(e)))
        raise e
    except Exception as e:
        print("error occured {}".format(str(e)))
        raise HTTPException(status_code=500, detail=str(e))


@app.put("/workers/pool-size")
async def repopulate(num_procs: int):
    pool_instance = TaskManager()
    if pool_instance.change_pool_size(num_procs):
        return "succeed"
    else:
        raise HTTPException(status_code=500, detail="repopulate internal error")


@app.get("/status")
async def heartbeat():
    return "success"


if __name__ == "__main__":
    uvicorn.run(app, host="127.0.0.1", port=9720)



import pyarrow as pa
import parse
import json
from stonewave.sql.udtfs.base_function import BaseFunction, udtf


@udtf(is_parser=True)
class ParseFormatFunction(BaseFunction):
    def __init__(self):
        self.compiled_format_cache = {}

    def get_name(self):
        return "parse_format"

    def process(self, row_writer, row_idx, args):
        text = args[0]
        pattern_format = args[1]
        compiled_format = None
        if pattern_format in self.compiled_format_cache:
            compiled_format = self.compiled_format_cache[pattern_format]
        else:
            compiled_format = parse.compile(pattern_format)
            self.compiled_format_cache[pattern_format] = compiled_format
        extracted_values = compiled_format.parse(text)

        record_batch_builder = row_writer.record_batch_builder
        if extracted_values:
            for col_name in extracted_values.named:
                record_batch_builder.add_column(col_name, pa.utf8())
                val = extracted_values[col_name]
                if isinstance(val, dict):
                    val = json.dumps(val)
                record_batch_builder.append(col_name, val)

            i = 0
            for v in extracted_values:
                col_name = "a" + str(i)
                record_batch_builder.add_column(col_name, pa.utf8())
                i += 1
                record_batch_builder.append(col_name, v)
            record_batch_builder.increase_row_count()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['stonewave',
 'stonewave.sql.udtfs',
 'stonewave.sql.udtfs.functions.faker',
 'stonewave.sql.udtfs.functions.generate_union',
 'stonewave.sql.udtfs.functions.load_excel',
 'stonewave.sql.udtfs.functions.parse_binary',
 'stonewave.sql.udtfs.functions.parse_binary.formats',
 'stonewave.sql.udtfs.functions.parse_format',
 'stonewave.sql.udtfs.functions.parse_grok',
 'stonewave.sql.udtfs.functions.parse_sql',
 'stonewave.sql.udtfs.functions.pivot_table',
 'stonewave.sql.udtfs.functions.summarize',
 'stonewave.sql.udtfs.functions.transpose',
 'stonewave.sql.udtfs.functions.unpivot_table',
 'stonewave.sql.udtfs.protocol',
 'stonewave.sql.udtfs.protocol.arrow',
 'stonewave.sql.udtfs.protocol.fsm',
 'stonewave.sql.udtfs.protocol.fsm.apply']

package_data = \
{'': ['*']}

install_requires = \
['faker-cloud>=0.1,<0.2',
 'faker>=5.0.2,<6.0.0',
 'faker_credit_score>=0.3.0,<0.4.0',
 'faker_microservice>=2.0.0,<3.0.0',
 'faker_vehicle>=0.1.3,<0.2.0',
 'faker_web>=0.3.1,<0.4.0',
 'faker_wifi_essid>=0.2.0,<0.3.0',
 'fastapi==0.70.0',
 'httpx>=0.20.0,<0.21.0',
 'kaitaistruct>=0.9,<0.10',
 'openpyxl>=3.0.6,<4.0.0',
 'pandas==1.2.5',
 'parse>=1.18.0,<2.0.0',
 'pyarrow==6.0.0',
 'pygrok>=1.0.0,<2.0.0',
 'requests>=2.25.1,<3.0.0',
 'setuptools>=41.2.0',
 'sql-metadata-fork>=2.3.0,<3.0.0',
 'structlog>=20.2.0,<21.0.0',
 'toml>=0.10.2,<0.11.0',
 'transitions>=0.8.3,<0.9.0',
 'trio>=0.19.0,<0.20.0',
 'uvicorn>=0.15.0,<0.16.0',
 'wheel-filename>=1.3.0,<2.0.0',
 'wheel>=0.35.1']

entry_points = \
{'console_scripts': ['sw_py_udtf = stonewave.sql.udtfs.main:execute_command']}

setup_kwargs = {
    'name': 'stonewave-sql-udtfs',
    'version': '0.5.1',
    'description': '',
    'long_description': None,
    'author': 'Yue Ni',
    'author_email': 'yni@yanhuangdata.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://www.yanhuangdata.com/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



from stonewave.sql.udtfs.logger import logger
from stonewave.sql.udtfs.base_function import BaseFunction
from datetime import datetime, timezone


class GenerateUnionFunction(BaseFunction):
    data_types = [
        ("null", lambda i: None),
        ("int", lambda i: i),
        ("float", lambda i: float(i)),
        ("bool", lambda i: bool(i)),
        ("string", lambda i: str(i)),
        ("timestamp", lambda i: datetime.fromtimestamp(i / 1e6, timezone.utc)),
        ("list<string>", lambda i: [str(i)]),
    ]

    def __init__(self):
        pass

    def get_name(self):
        return "generate_union"

    def process(self, row_writer, row_idx, args):
        rng = range(0, args[0])
        types = [i.strip() for i in args[1].split(",")]
        array = []
        for idx, data_type in enumerate(GenerateUnionFunction.data_types):
            parsed_indices = []
            for idx_ts, type_str in enumerate(types):
                if type_str == data_type[0]:
                    array.extend([data_type[1](i) for i in rng])
                    parsed_indices.append(idx_ts)
            for i in reversed(parsed_indices):
                types.pop(i)
        record_batch_builder = row_writer.record_batch_builder
        col_name = "generate_union"
        record_batch_builder.add_column(col_name)
        record_batch_builder.extend(col_name, array)
        row_count = len(array)
        record_batch_builder.increase_row_count(row_count)
        logger.debug(
            "rows generated via generate_union",
            row_count=row_count,
            builder_row_count=record_batch_builder.row_count(),
        )



from stonewave.sql.udtfs.base_function import BaseFunction, udtf
import importlib


# ssh_public_key ==> SshPublicKey
def _to_camel_case(snake_str):
    components = snake_str.split("_")
    return "".join([c.capitalize() for c in components])


# TODO: how can we combine with the arrow dataset with this function?


@udtf(is_parser=True)
class ParseBinaryFunction(BaseFunction):
    def __init__(self):
        pass

    def get_name(self):
        return "parse_binary"

    def process(self, row_writer, row_idx, args):
        file_path = args[0]
        format_name = args[1]
        data_paths = [c.strip() for c in args[2].split(",")] if len(args) > 2 else []
        kaitai_parser = "stonewave.sql.udtfs.functions.parse_binary.formats.{}".format(format_name)
        module = importlib.import_module(kaitai_parser)
        parser_class_name = _to_camel_case(format_name)
        parser_class = getattr(module, parser_class_name)
        # TODO: there is some security concern, we probably need to limit the path it can access
        parsed_binary = parser_class.from_file(file_path)
        kv_pairs = {}
        for path in data_paths:
            # TODO: can we marshall it into a JSON string?
            value = str(getattr(parsed_binary, path))
            kv_pairs[path] = value
        row_writer.record_batch_builder.append_row(kv_pairs)
